import json
from typing import List
from clarifai_grpc.grpc.api import resources_pb2
from clarifai.runners.utils import data_types


def generate_client_script(method_signatures: List[resources_pb2.MethodSignature], user_id, app_id,
                           model_id) -> str:
  _CLIENT_TEMPLATE = """\
from clarifai.runners.utils import data_types
import os
from clarifai.client import Model

model = Model("www.clarifai.com/{user_id}/{app_id}/{model_id}",
    deployment_id = os.environ['CLARIFAI_DEPLOYMENT_ID'], # Only needed for dedicated deployed models
)
    """
  # Generate client template
  client_template = _CLIENT_TEMPLATE.format(user_id=user_id, app_id=app_id, model_id=model_id)

  # Generate method signatures
  method_signatures_str = []
  for method_signature in method_signatures:
    method_name = method_signature.name
    if method_signature.method_type in [
        resources_pb2.RunnerMethodType.UNARY_UNARY,
        resources_pb2.RunnerMethodType.UNARY_STREAMING,
    ]:

      client_script_str = f'response = model.{method_name}('
      annotations = _get_annotations_source(method_signature)
      for param_name, (param_type, default_value) in annotations.items():
        if param_name == "return":
          continue
        if default_value is None:
          default_value = _set_default_value(param_type)
          if param_type == "str":
            default_value = repr(default_value)

        client_script_str += f"{param_name}={default_value}, "
      client_script_str = client_script_str.rstrip(", ") + ")"
      if method_signature.method_type == resources_pb2.RunnerMethodType.UNARY_UNARY:
        client_script_str += "\nprint(response)"
      elif method_signature.method_type == resources_pb2.RunnerMethodType.UNARY_STREAMING:
        client_script_str += "\nfor res in response:\n    print(res)"
      client_script_str += "\n"
      method_signatures_str.append(client_script_str)

  method_signatures_str = "\n".join(method_signatures_str)
  # Combine all parts
  script_lines = []
  script_lines.append("\n# Clarifai Model Client Script")
  script_lines.append(
      "# Set the environment variables `CLARIFAI_DEPLOYMENT_ID` and `CLARIFAI_PAT` to run this script."
  )
  script_lines.append("# Example usage:")
  script_lines.append(client_template)
  script_lines.append("# Example model prediction from different model methods: \n")
  script_lines.append(method_signatures_str)
  script_lines.append("")
  script = "\n".join(script_lines)
  return script


# get annotations source with default values
def _get_annotations_source(method_signature: resources_pb2.MethodSignature) -> dict:
  annotations = {}
  for input_field in method_signature.input_fields:
    param_name = input_field.name
    param_type = _get_base_type(input_field)
    if input_field.iterator:
      param_type = f"Iterator[{param_type}]"
    default_value = None
    if input_field.default:
      default_value = _parse_default_value(input_field)

    annotations[param_name] = (param_type, default_value)
  if not method_signature.output_fields:
    raise ValueError("MethodSignature must have at least one output field")
  for output_field in method_signature.output_fields:
    param_name = output_field.name
    param_type = _get_base_type(output_field)
    if output_field.iterator:
      param_type = f"Iterator[{param_type}]"
    annotations[param_name] = (param_type, None)
  return annotations


def _get_base_type(field: resources_pb2.ModelTypeField) -> str:
  data_type = field.type
  if data_type == resources_pb2.ModelTypeField.DataType.NAMED_FIELDS:
    annotations = {}
    for type_arg in field.type_args:
      field_name = type_arg.name
      field_type = _get_base_type(type_arg)
      annotations[field_name] = field_type
    class_name = f"NamedFields[{', '.join(f'{k}: {v}' for k, v in annotations.items())}]"
    return class_name
  elif data_type == resources_pb2.ModelTypeField.DataType.TUPLE:
    type_args_str = [_get_base_type(ta) for ta in field.type_args]
    return f"Tuple[{', '.join(type_args_str)}]"
  elif data_type == resources_pb2.ModelTypeField.DataType.LIST:
    if len(field.type_args) != 1:
      raise ValueError("List type must have exactly one type argument")
    element_type = _get_base_type(field.type_args[0])
    return f"List[{element_type}]"
  else:
    type_map = {
        resources_pb2.ModelTypeField.DataType.STR: "str",
        resources_pb2.ModelTypeField.DataType.BYTES: "bytes",
        resources_pb2.ModelTypeField.DataType.INT: "int",
        resources_pb2.ModelTypeField.DataType.FLOAT: "float",
        resources_pb2.ModelTypeField.DataType.BOOL: "bool",
        resources_pb2.ModelTypeField.DataType.NDARRAY: "np.ndarray",
        resources_pb2.ModelTypeField.DataType.JSON_DATA: "data_types.JSON",
        resources_pb2.ModelTypeField.DataType.TEXT: "data_types.Text",
        resources_pb2.ModelTypeField.DataType.IMAGE: "data_types.Image",
        resources_pb2.ModelTypeField.DataType.CONCEPT: "data_types.Concept",
        resources_pb2.ModelTypeField.DataType.REGION: "data_types.Region",
        resources_pb2.ModelTypeField.DataType.FRAME: "data_types.Frame",
        resources_pb2.ModelTypeField.DataType.AUDIO: "data_types.Audio",
        resources_pb2.ModelTypeField.DataType.VIDEO: "data_types.Video",
    }
    return type_map.get(data_type, "Any")


def _map_default_value(field_type):
  """
    Map the default value of a field type to a string representation.
    """
  default_value = None

  if field_type == "str":
    default_value = 'What is the future of AI?'
  elif field_type == "bytes":
    default_value = b""
  elif field_type == "int":
    default_value = 0
  elif field_type == "float":
    default_value = 0.0
  elif field_type == "bool":
    default_value = False
  elif field_type == "data_types.Image":
    default_value = data_types.Image.from_url("https://samples.clarifai.com/metro-north.jpg")
  elif field_type == "data_types.Text":
    default_value = data_types.Text("What's the future of AI?")
  elif field_type == "data_types.Audio":
    default_value = data_types.Audio.from_url("https://samples.clarifai.com/audio.mp3")
  elif field_type == "data_types.Video":
    default_value = data_types.Video.from_url("https://samples.clarifai.com/video.mp4")
  elif field_type == "data_types.Concept":
    default_value = data_types.Concept(id="concept_id", name="dog", value=0.95)
  elif field_type == "data_types.Region":
    default_value = data_types.Region(box=[0.1, 0.1, 0.5, 0.5],)
  elif field_type == "data_types.Frame":
    default_value = data_types.Frame.from_url("https://samples.clarifai.com/video.mp4", 0)
  elif field_type == "data_types.NDArray":
    default_value = data_types.NDArray([1, 2, 3])
  else:
    default_value = None
  return default_value


def _set_default_value(field_type):
  """
    Set the default value of a field if it is not set.
    """
  default_value = None
  default_value = _map_default_value(field_type)
  if field_type.startswith("List["):
    element_type = field_type[5:-1]
    element_type_default_value = _map_default_value(element_type)
    if element_type_default_value is not None:
      default_value = f"[{element_type_default_value}]"
  elif field_type.startswith("Tuple["):
    element_types = field_type[6:-1].split(", ")
    element_type_defaults = [_map_default_value(et) for et in element_types]
    default_value = f"({', '.join([str(et) for et in element_type_defaults])})"
  elif field_type.startswith("NamedFields["):
    element_types = field_type[12:-1].split(", ")
    element_type_defaults = [_map_default_value(et) for et in element_types]
    default_value = f"{{{', '.join([str(et) for et in element_type_defaults])}}}"

  return default_value


def _parse_default_value(field: resources_pb2.ModelTypeField):
  if not field.default:
    return None
  default_str = field.default
  data_type = field.type

  try:
    if data_type == resources_pb2.ModelTypeField.DataType.INT:
      return str(int(default_str))
    elif data_type == resources_pb2.ModelTypeField.DataType.FLOAT:
      return str(float(default_str))
    elif data_type == resources_pb2.ModelTypeField.DataType.BOOL:
      return 'True' if default_str.lower() == 'true' else 'False'
    elif data_type == resources_pb2.ModelTypeField.DataType.STR:
      return repr(default_str)
    elif data_type == resources_pb2.ModelTypeField.DataType.BYTES:
      return f"b{repr(default_str.encode('utf-8'))}"
    elif data_type == resources_pb2.ModelTypeField.DataType.JSON_DATA:
      parsed = json.loads(default_str)
      return repr(parsed)
    elif data_type in (
        resources_pb2.ModelTypeField.DataType.LIST,
        resources_pb2.ModelTypeField.DataType.TUPLE,
        resources_pb2.ModelTypeField.DataType.NAMED_FIELDS,
    ):
      parsed = json.loads(default_str)
      return repr(parsed)
    else:
      return repr(default_str)
  except (ValueError, json.JSONDecodeError):
    return repr(default_str)



import inspect
import time
from typing import Any, Dict, Iterator, List

from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2

from clarifai.constants.model import MAX_MODEL_PREDICT_INPUTS
from clarifai.errors import UserError
from clarifai.runners.utils import code_script, method_signatures
from clarifai.runners.utils.data_utils import is_openai_chat_format
from clarifai.runners.utils.method_signatures import (CompatibilitySerializer, deserialize,
                                                      get_stream_from_signature, serialize,
                                                      signatures_from_json)
from clarifai.utils.logging import logger
from clarifai.utils.misc import BackoffIterator, status_is_retryable


class ModelClient:
  '''
  Client for calling model predict, generate, and stream methods.
  '''

  def __init__(self, stub, request_template: service_pb2.PostModelOutputsRequest = None):
    '''
        Initialize the model client.

        Args:
            stub: The gRPC stub for the model.
            request_template: The template for the request to send to the model, including
            common fields like model_id, model_version, cluster, etc.
        '''
    self.STUB = stub
    self.request_template = request_template or service_pb2.PostModelOutputsRequest()
    self._method_signatures = None
    self._defined = False

  def fetch(self):
    '''
    Fetch function signature definitions from the model and define the functions in the client
    '''
    if self._defined:
      return
    try:
      self._fetch_signatures()
      self._define_functions()
    finally:
      self._defined = True

  def __getattr__(self, name):
    if not self._defined:
      self.fetch()
    return self.__getattribute__(name)

  def _fetch_signatures(self):
    '''
      Fetch the method signatures from the model.

      Returns:
          Dict: The method signatures.
      '''
    try:
      response = self.client.STUB.GetModelVersion(
          service_pb2.GetModelVersionRequest(
              user_app_id=self.request_template.user_app_id,
              model_id= self.request_template.model_id,
              version_id=self.request_template.model_version.id,
          ))
      method_signatures = None
      if response.status.code == status_code_pb2.SUCCESS:
        method_signatures = response.model_version.method_signatures
      if response.status.code != status_code_pb2.SUCCESS:
        raise Exception(f"Model failed with response {response!r}")
      self._method_signatures= {}
      for method_signature in method_signatures:
        method_name = method_signature.name
        # check for duplicate method names
        if method_name in self._method_signatures:
          raise ValueError(f"Duplicate method name {method_name}")
        self._method_signatures[method_name] = method_signature
      if not self._method_signatures: # if no method signatures, try to fetch from the model
        self._fetch_signatures_backup()
    except Exception as e:
      logger.info(f"Failed to fetch method signatures from model: {e}")
      # try to fetch from the model
      self._fetch_signatures_backup()
      if not self._method_signatures:
        raise ValueError("Failed to fetch method signatures from model and backup method")

  def _fetch_signatures_backup(self):
    '''
      This is a temporary method of fetching the method signatures from the model.

      Returns:
          Dict: The method signatures.
      '''

    request = service_pb2.PostModelOutputsRequest()
    request.CopyFrom(self.request_template)
    # request.model.model_version.output_info.params['_method_name'] = '_GET_SIGNATURES'
    inp = request.inputs.add()  # empty input for this method
    inp.data.parts.add()  # empty part for this input
    inp.data.metadata['_method_name'] = '_GET_SIGNATURES'
    start_time = time.time()
    backoff_iterator = BackoffIterator(10)
    while True:
      response = self.STUB.PostModelOutputs(request)
      if status_is_retryable(
          response.status.code) and time.time() - start_time < 60 * 10:  # 10 minutes
        logger.info(f"Retrying model info fetch with response {response.status!r}")
        time.sleep(next(backoff_iterator))
        continue
      break
    if (response.status.code == status_code_pb2.INPUT_UNSUPPORTED_FORMAT or
        (response.status.code == status_code_pb2.SUCCESS and
         response.outputs[0].data.text.raw == '')):
      # return codes/values from older models that don't support _GET_SIGNATURES
      self._method_signatures = {}
      self._define_compatability_functions()
      return
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(f"Model failed with response {response!r}")
    self._method_signatures = signatures_from_json(response.outputs[0].data.text.raw)

  def _define_functions(self):
    '''
    Define the functions based on the method signatures.
    '''
    for method_name, method_signature in self._method_signatures.items():
      # define the function in this client instance
      if resources_pb2.RunnerMethodType.Name(method_signature.method_type) == 'UNARY_UNARY':
        call_func = self._predict
      elif resources_pb2.RunnerMethodType.Name(method_signature.method_type) == 'UNARY_STREAMING':
        call_func = self._generate
      elif resources_pb2.RunnerMethodType.Name(
          method_signature.method_type) == 'STREAMING_STREAMING':
        call_func = self._stream
      else:
        raise ValueError(f"Unknown method type {method_signature.method_type}")

      # method argnames, in order, collapsing nested keys to corresponding user function args
      method_argnames = []
      for var in method_signature.input_fields:
        outer = var.name.split('.', 1)[0]
        if outer in method_argnames:
          continue
        method_argnames.append(outer)

      def bind_f(method_name, method_argnames, call_func):

        def f(*args, **kwargs):
          if len(args) > len(method_argnames):
            raise TypeError(
                f"{method_name}() takes {len(method_argnames)} positional arguments but {len(args)} were given"
            )

          if len(args) + len(kwargs) > len(method_argnames):
            raise TypeError(
                f"{method_name}() got an unexpected keyword argument {next(iter(kwargs))}")
          if len(args) == 1 and (not kwargs) and isinstance(args[0], list):
            batch_inputs = args[0]
            # Validate each input is a dictionary
            is_batch_input_valid = all(isinstance(input, dict) for input in batch_inputs)
            if is_batch_input_valid and (not is_openai_chat_format(batch_inputs)):
              # If the batch input is valid, call the function with the batch inputs and the method name
              return call_func(batch_inputs, method_name)

          for name, arg in zip(method_argnames, args):  # handle positional with zip shortest
            if name in kwargs:
              raise TypeError(f"Multiple values for argument {name}")
            kwargs[name] = arg
          return call_func(kwargs, method_name)

        return f

      # need to bind method_name to the value, not the mutating loop variable
      f = bind_f(method_name, method_argnames, call_func)

      # set names, annotations and docstrings
      f.__name__ = method_name
      f.__qualname__ = f'{self.__class__.__name__}.{method_name}'
      input_annotations = code_script._get_annotations_source(method_signature)
      return_annotation = input_annotations.pop('return', (None, None))[0]
      sig = inspect.signature(f).replace(
          parameters=[
              inspect.Parameter(k, inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=v[0])
              for k, v in input_annotations.items()
          ],
          return_annotation=return_annotation,
      )
      f.__signature__ = sig
      f.__doc__ = method_signature.description
      setattr(self, method_name, f)

  def available_methods(self) -> List[str]:
    """Get the available methods for this model.

    Returns:
        List[str]: The available methods.
    """
    if not self._defined:
      self.fetch()
    return self._method_signatures.keys()

  def method_signature(self, method_name: str) -> str:
    """Get the method signature for a method.

    Args:
        method_name (str): The name of the method.

    Returns:
        str: The method signature.
    """
    if not self._defined:
      self.fetch()
    return method_signatures.get_method_signature(self._method_signatures[method_name])

  def generate_client_script(self) -> str:
    """Generate a client script for this model.

    Returns:
        str: The client script.
    """
    if not self._defined:
      self.fetch()
    method_signatures = []
    for _, method_signature in self._method_signatures.items():
      method_signatures.append(method_signature)
    return code_script.generate_client_script(
        method_signatures,
        user_id=self.request_template.user_app_id.user_id,
        app_id=self.request_template.user_app_id.app_id,
        model_id=self.request_template.model_id)

  def _define_compatability_functions(self):

    serializer = CompatibilitySerializer()

    def predict(input: Any) -> Any:
      proto = resources_pb2.Input()
      serializer.serialize(proto.data, input)
      # always use text.raw for compat
      if proto.data.string_value:
        proto.data.text.raw = proto.data.string_value
        proto.data.string_value = ''
      response = self._predict_by_proto([proto])
      if response.status.code != status_code_pb2.SUCCESS:
        raise Exception(f"Model predict failed with response {response!r}")
      response_data = response.outputs[0].data
      if response_data.text.raw:
        response_data.string_value = response_data.text.raw
        response_data.text.raw = ''
      return serializer.deserialize(response_data)

    self.predict = predict

  def _predict(
      self,
      inputs,  # TODO set up functions according to fetched signatures?
      method_name: str = 'predict',
  ) -> Any:

    input_signature = self._method_signatures[method_name].input_fields
    output_signature = self._method_signatures[method_name].output_fields

    batch_input = True
    if isinstance(inputs, dict):
      inputs = [inputs]
      batch_input = False

    proto_inputs = []
    for input in inputs:
      proto = resources_pb2.Input()

      serialize(input, input_signature, proto.data)
      proto_inputs.append(proto)

    response = self._predict_by_proto(proto_inputs, method_name)

    outputs = []
    for output in response.outputs:
      outputs.append(deserialize(output.data, output_signature, is_output=True))
    if batch_input:
      return outputs
    return outputs[0]

  def _predict_by_proto(
      self,
      inputs: List[resources_pb2.Input],
      method_name: str = None,
      inference_params: Dict = None,
      output_config: Dict = None,
  ) -> service_pb2.MultiOutputResponse:
    """Predicts the model based on the given inputs.

      Args:
          inputs (List[resources_pb2.Input]): The inputs to predict.
          method_name (str): The remote method name to call.
          inference_params (Dict): Inference parameters to override.
          output_config (Dict): Output configuration to override.

      Returns:
          service_pb2.MultiOutputResponse: The prediction response(s).
      """
    if not isinstance(inputs, list):
      raise UserError('Invalid inputs, inputs must be a list of Input objects.')
    if len(inputs) > MAX_MODEL_PREDICT_INPUTS:
      raise UserError(f"Too many inputs. Max is {MAX_MODEL_PREDICT_INPUTS}.")

    request = service_pb2.PostModelOutputsRequest()
    request.CopyFrom(self.request_template)

    request.inputs.extend(inputs)

    if method_name:
      # TODO put in new proto field?
      for inp in request.inputs:
        inp.data.metadata['_method_name'] = method_name
    if inference_params:
      request.model.model_version.output_info.params.update(inference_params)
    if output_config:
      request.model.model_version.output_info.output_config.MergeFrom(
          resources_pb2.OutputConfig(**output_config))

    start_time = time.time()
    backoff_iterator = BackoffIterator(10)
    while True:
      response = self.STUB.PostModelOutputs(request)
      if status_is_retryable(
          response.status.code) and time.time() - start_time < 60 * 10:  # 10 minutes
        logger.info(f"Model predict failed with response {response!r}")
        time.sleep(next(backoff_iterator))
        continue

      if response.status.code != status_code_pb2.SUCCESS:
        raise Exception(f"Model predict failed with response {response!r}")
      break
    return response

  def _generate(
      self,
      inputs,  # TODO set up functions according to fetched signatures?
      method_name: str = 'generate',
  ) -> Any:
    input_signature = self._method_signatures[method_name].input_fields
    output_signature = self._method_signatures[method_name].output_fields

    batch_input = True
    if isinstance(inputs, dict):
      inputs = [inputs]
      batch_input = False

    proto_inputs = []
    for input in inputs:
      proto = resources_pb2.Input()
      serialize(input, input_signature, proto.data)
      proto_inputs.append(proto)

    response_stream = self._generate_by_proto(proto_inputs, method_name)

    for response in response_stream:
      outputs = []
      for output in response.outputs:
        outputs.append(deserialize(output.data, output_signature, is_output=True))
      if batch_input:
        yield outputs
      else:
        yield outputs[0]

  def _generate_by_proto(
      self,
      inputs: List[resources_pb2.Input],
      method_name: str = None,
      inference_params: Dict = {},
      output_config: Dict = {},
  ):
    """Generate the stream output on model based on the given inputs.

    Args:
        inputs (list[Input]): The inputs to generate, must be less than 128.
        method_name (str): The remote method name to call.
        inference_params (dict): The inference params to override.
        output_config (dict): The output config to override.
    """
    if not isinstance(inputs, list):
      raise UserError('Invalid inputs, inputs must be a list of Input objects.')
    if len(inputs) > MAX_MODEL_PREDICT_INPUTS:
      raise UserError(f"Too many inputs. Max is {MAX_MODEL_PREDICT_INPUTS}."
                     )  # TODO Use Chunker for inputs len > 128

    request = service_pb2.PostModelOutputsRequest()
    request.CopyFrom(self.request_template)

    request.inputs.extend(inputs)

    if method_name:
      # TODO put in new proto field?
      for inp in request.inputs:
        inp.data.metadata['_method_name'] = method_name
    if inference_params:
      request.model.model_version.output_info.params.update(inference_params)
    if output_config:
      request.model.model_version.output_info.output_config.MergeFromDict(output_config)

    start_time = time.time()
    backoff_iterator = BackoffIterator(10)
    started = False
    while not started:
      stream_response = self.STUB.GenerateModelOutputs(request)
      try:
        response = next(stream_response)  # get the first response
      except StopIteration:
        raise Exception("Model Generate failed with no response")
      if status_is_retryable(response.status.code) and \
              time.time() - start_time < 60 * 10:
        logger.info("Model is still deploying, please wait...")
        time.sleep(next(backoff_iterator))
        continue
      if response.status.code != status_code_pb2.SUCCESS:
        raise Exception(f"Model Generate failed with response {response.status!r}")
      started = True

    yield response  # yield the first response

    for response in stream_response:
      if response.status.code != status_code_pb2.SUCCESS:
        raise Exception(f"Model Generate failed with response {response.status!r}")
      yield response

  def _stream(
      self,
      inputs,
      method_name: str = 'stream',
  ) -> Any:
    input_signature = self._method_signatures[method_name].input_fields
    output_signature = self._method_signatures[method_name].output_fields

    if isinstance(inputs, list):
      assert len(inputs) == 1, 'streaming methods do not support batched calls'
      inputs = inputs[0]
    assert isinstance(inputs, dict)
    kwargs = inputs

    # find the streaming vars in the input signature, and the streaming input python param
    stream_sig = get_stream_from_signature(input_signature)
    if stream_sig is None:
      raise ValueError("Streaming method must have a Stream input")
    stream_argname = stream_sig.name

    # get the streaming input generator from the user-provided function arg values
    user_inputs_generator = kwargs.pop(stream_argname)

    def _input_proto_stream():
      # first item contains all the inputs and the first stream item
      proto = resources_pb2.Input()
      try:
        item = next(user_inputs_generator)
      except StopIteration:
        return  # no items to stream
      kwargs[stream_argname] = item
      serialize(kwargs, input_signature, proto.data)

      yield proto

      # subsequent items are just the stream items
      for item in user_inputs_generator:
        proto = resources_pb2.Input()
        serialize({stream_argname: item}, [stream_sig], proto.data)
        yield proto

    response_stream = self._stream_by_proto(_input_proto_stream(), method_name)

    for response in response_stream:
      assert len(response.outputs) == 1, 'streaming methods must have exactly one output'
      yield deserialize(response.outputs[0].data, output_signature, is_output=True)

  def _req_iterator(self,
                    input_iterator: Iterator[List[resources_pb2.Input]],
                    method_name: str = None,
                    inference_params: Dict = {},
                    output_config: Dict = {}):
    request = service_pb2.PostModelOutputsRequest()
    request.CopyFrom(self.request_template)
    if inference_params:
      request.model.model_version.output_info.params.update(inference_params)
    if output_config:
      request.model.model_version.output_info.output_config.MergeFromDict(output_config)
    for inputs in input_iterator:
      req = service_pb2.PostModelOutputsRequest()
      req.CopyFrom(request)
      if isinstance(inputs, list):
        req.inputs.extend(inputs)
      else:
        req.inputs.append(inputs)
      # TODO: put into new proto field?
      if method_name:
        for inp in req.inputs:
          inp.data.metadata['_method_name'] = method_name
      yield req

  def _stream_by_proto(self,
                       inputs: Iterator[List[resources_pb2.Input]],
                       method_name: str = None,
                       inference_params: Dict = {},
                       output_config: Dict = {}):
    """Generate the stream output on model based on the given stream of inputs.
    """
    # if not isinstance(inputs, Iterator[List[Input]]):
    #   raise UserError('Invalid inputs, inputs must be a iterator of list of Input objects.')

    request = self._req_iterator(inputs, method_name, inference_params, output_config)

    start_time = time.time()
    backoff_iterator = BackoffIterator(10)
    generation_started = False
    while True:
      if generation_started:
        break
      stream_response = self.STUB.StreamModelOutputs(request)
      for response in stream_response:
        if status_is_retryable(response.status.code) and \
                time.time() - start_time < 60 * 10:
          logger.info("Model is still deploying, please wait...")
          time.sleep(next(backoff_iterator))
          break
        if response.status.code != status_code_pb2.SUCCESS:
          raise Exception(f"Model Predict failed with response {response.status!r}")
        else:
          if not generation_started:
            generation_started = True
          yield response



import collections.abc as abc
import inspect
import json
from collections import namedtuple
from typing import Dict, List, Tuple, get_args, get_origin

import numpy as np
import PIL.Image
import yaml
from clarifai_grpc.grpc.api import resources_pb2
from google.protobuf.json_format import MessageToDict, ParseDict
from google.protobuf.message import Message as MessageProto

from clarifai.runners.utils import data_types, data_utils
from clarifai.runners.utils.code_script import _get_base_type, _parse_default_value
from clarifai.runners.utils.serializers import (
    AtomicFieldSerializer, JSONSerializer, ListSerializer, MessageSerializer,
    NamedFieldsSerializer, NDArraySerializer, Serializer, TupleSerializer)


def build_function_signature(func):
  '''
  Build a signature for the given function.
  '''
  sig = inspect.signature(func)

  # check if func is bound, and if not, remove self/cls
  if getattr(func, '__self__', None) is None and sig.parameters and list(
      sig.parameters.values())[0].name in ('self', 'cls'):
    sig = sig.replace(parameters=list(sig.parameters.values())[1:])

  return_annotation = sig.return_annotation
  if return_annotation == inspect.Parameter.empty:
    raise TypeError('Function must have a return annotation')

  input_sigs = []
  input_streaming = []
  for p in sig.parameters.values():
    model_type_field, _, streaming = build_variable_signature(p.name, p.annotation, p.default)
    input_sigs.append(model_type_field)
    input_streaming.append(streaming)

  output_sig, output_type, output_streaming = build_variable_signature(
      'return', return_annotation, is_output=True)
  # TODO: flatten out "return" layer if not needed

  # check for streams and determine method type
  if sum(input_streaming) > 1:
    raise TypeError('streaming methods must have at most one streaming input')
  input_streaming = any(input_streaming)
  if not (input_streaming or output_streaming):
    method_type = 'UNARY_UNARY'
  elif not input_streaming and output_streaming:
    method_type = 'UNARY_STREAMING'
  elif input_streaming and output_streaming:
    method_type = 'STREAMING_STREAMING'
  else:
    raise TypeError('stream methods with streaming inputs must have streaming outputs')

  method_signature = resources_pb2.MethodSignature()

  method_signature.name = func.__name__
  method_signature.method_type = getattr(resources_pb2.RunnerMethodType, method_type)
  assert method_type in ('UNARY_UNARY', 'UNARY_STREAMING', 'STREAMING_STREAMING')
  # method_signature.method_type = method_type
  method_signature.description = inspect.cleandoc(func.__doc__ or '')
  # method_signature.annotations_json = json.dumps(_get_annotations_source(func))

  method_signature.input_fields.extend(input_sigs)
  method_signature.output_fields.append(output_sig)
  return method_signature


# def _get_annotations_source(func):
#   """Extracts raw annotation strings from the function source."""
#   source = inspect.getsource(func)  # Get function source code
#   source = textwrap.dedent(source)  # Dedent source code
#   tree = ast.parse(source)  # Parse into AST
#   func_node = next(node for node in tree.body
#                    if isinstance(node, ast.FunctionDef))  # Get function node

#   annotations = {}
#   for arg in func_node.args.args:  # Process arguments
#     if arg.annotation:
#       annotations[arg.arg] = ast.unparse(arg.annotation)  # Get raw annotation string

#   if func_node.returns:  # Process return type
#     annotations["return"] = ast.unparse(func_node.returns)

#   return annotations


def _process_input_field(field: resources_pb2.ModelTypeField) -> str:
  base_type = _get_base_type(field)
  if field.iterator:
    type_str = f"Iterator[{base_type}]"
  else:
    type_str = base_type
  default = _parse_default_value(field)
  param = f"{field.name}: {type_str}"
  if default is not None:
    param += f" = {default}"
  return param


def _process_output_field(field: resources_pb2.ModelTypeField) -> str:
  base_type = _get_base_type(field)
  if field.iterator:
    return f"Iterator[{base_type}]"
  else:
    return base_type


def get_method_signature(method_signature: resources_pb2.MethodSignature) -> str:
  """
    Get the method signature of a method in a model.
    """
  # Process input fields
  input_params = []
  for input_field in method_signature.input_fields:
    param_str = _process_input_field(input_field)
    input_params.append(param_str)

  # Process output field
  if not method_signature.output_fields:
    raise ValueError("MethodSignature must have at least one output field")
  output_field = method_signature.output_fields[0]
  return_type = _process_output_field(output_field)

  # Generate function signature
  function_def = f"def {method_signature.name}({', '.join(input_params)}) -> {return_type}:"
  return function_def


def build_variable_signature(name, annotation, default=inspect.Parameter.empty, is_output=False):
  '''
  Build a data proto signature and get the normalized python type for the given annotation.
  '''

  # check valid names (should already be constrained by python naming, but check anyway)
  if not name.isidentifier():
    raise ValueError(f'Invalid variable name: {name}')

  # get fields for each variable based on type
  tp, streaming = _normalize_type(annotation)

  sig = resources_pb2.ModelTypeField()
  sig.name = name
  sig.iterator = streaming

  if not is_output:
    sig.required = (default is inspect.Parameter.empty)
    if not sig.required:
      if isinstance(default, data_utils.InputField):
        sig = default.to_proto(sig)
      else:
        sig = data_utils.InputField.set_default(sig, default)

  _fill_signature_type(sig, tp)

  return sig, type, streaming


def _fill_signature_type(sig, tp):
  try:
    if tp in _DATA_TYPES:
      sig.type = _DATA_TYPES[tp].type
      return
  except TypeError:
    pass  # not hashable type

  # Handle NamedFields with annotations
  # Check for dynamically generated NamedFields subclasses (from type annotations)
  if inspect.isclass(tp) and issubclass(tp, data_types.NamedFields) and hasattr(
      tp, '__annotations__'):
    sig.type = resources_pb2.ModelTypeField.DataType.NAMED_FIELDS
    for name, inner_type in tp.__annotations__.items():
      inner_sig = sig.type_args.add()
      inner_sig.name = name
      _fill_signature_type(inner_sig, inner_type)
    return

  # Handle NamedFields instances (dict-like)
  if isinstance(tp, data_types.NamedFields):
    sig.type = resources_pb2.ModelTypeField.DataType.NAMED_FIELDS
    for name, inner_type in tp.items():
      inner_sig = sig.type_args.add()
      inner_sig.name = name
      _fill_signature_type(inner_sig, inner_type)
    return

  origin = get_origin(tp)
  args = get_args(tp)

  # Handle Tuple type
  if origin == tuple:
    sig.type = resources_pb2.ModelTypeField.DataType.TUPLE
    for inner_type in args:
      inner_sig = sig.type_args.add()
      inner_sig.name = sig.name + '_item'
      _fill_signature_type(inner_sig, inner_type)
    return

  # Handle List type
  if origin == list:
    sig.type = resources_pb2.ModelTypeField.DataType.LIST
    inner_sig = sig.type_args.add()
    inner_sig.name = sig.name + '_item'
    _fill_signature_type(inner_sig, args[0])
    return

  raise TypeError(f'Unsupported type: {tp}')


def serializer_from_signature(signature):
  '''
    Get the serializer for the given signature.
    '''
  if signature.type in _SERIALIZERS_BY_TYPE_ENUM:
    return _SERIALIZERS_BY_TYPE_ENUM[signature.type]
  if signature.type == resources_pb2.ModelTypeField.DataType.LIST:
    return ListSerializer(serializer_from_signature(signature.type_args[0]))
  if signature.type == resources_pb2.ModelTypeField.DataType.TUPLE:
    return TupleSerializer([serializer_from_signature(sig) for sig in signature.type_args])
  if signature.type == resources_pb2.ModelTypeField.DataType.NAMED_FIELDS:
    return NamedFieldsSerializer(
        {sig.name: serializer_from_signature(sig)
         for sig in signature.type_args})
  raise ValueError(f'Unsupported type: {signature.type}')


def signatures_to_json(signatures):
  assert isinstance(
      signatures, dict), 'Expected dict of signatures {name: signature}, got %s' % type(signatures)
  # TODO change to proto when ready
  signatures = {name: MessageToDict(sig) for name, sig in signatures.items()}
  return json.dumps(signatures)


def signatures_from_json(json_str):
  signatures_dict = json.loads(json_str)
  assert isinstance(signatures_dict, dict), "Expected JSON to decode into a dictionary"

  return {
      name: ParseDict(sig_dict, resources_pb2.MethodSignature())
      for name, sig_dict in signatures_dict.items()
  }
  # d = json.loads(json_str, object_pairs_hook=_SignatureDict)
  # return d


def signatures_to_yaml(signatures):
  # XXX go in/out of json to get the correct format and python dict types
  d = json.loads(signatures_to_json(signatures))

  def _filter_empty(d):
    if isinstance(d, (list, tuple)):
      return [_filter_empty(v) for v in d if v]
    if isinstance(d, dict):
      return {k: _filter_empty(v) for k, v in d.items() if v}
    return d

  return yaml.dump(_filter_empty(d), default_flow_style=False)


def signatures_from_yaml(yaml_str):
  d = yaml.safe_load(yaml_str)
  return signatures_from_json(json.dumps(d))


def serialize(kwargs, signatures, proto=None, is_output=False):
  '''
  Serialize the given kwargs into the proto using the given signatures.
  '''
  if proto is None:
    proto = resources_pb2.Data()
  unknown = set(kwargs.keys()) - set(sig.name for sig in signatures)
  if unknown:
    if unknown == {'return'} and len(signatures) > 1:
      raise TypeError('Got a single return value, but expected multiple outputs {%s}' %
                      ', '.join(sig.name for sig in signatures))
    raise TypeError('Got unexpected key: %s' % ', '.join(unknown))
  for sig_i, sig in enumerate(signatures):
    if sig.name not in kwargs:
      if sig.required:
        raise TypeError(f'Missing required argument: {sig.name}')
      continue  # skip missing fields, they can be set to default on the server
    data = kwargs[sig.name]
    serializer = serializer_from_signature(sig)
    # TODO determine if any (esp the first) var can go in the proto without parts
    # and whether to put this in the signature or dynamically determine it
    # add the part to the proto
    part = proto.parts.add()
    part.id = sig.name
    serializer.serialize(part.data, data)
  return proto


def deserialize(proto, signatures, inference_params={}, is_output=False):
  '''
  Deserialize the given proto into kwargs using the given signatures.
  '''
  if isinstance(signatures, dict):
    signatures = [signatures]  # TODO update return key level and make consistnet
  kwargs = {}
  parts_by_name = {part.id: part for part in proto.parts}
  for sig_i, sig in enumerate(signatures):
    serializer = serializer_from_signature(sig)
    part = parts_by_name.get(sig.name)
    inference_params_value = inference_params.get(sig.name)
    if part is not None:
      kwargs[sig.name] = serializer.deserialize(part.data)
    elif inference_params_value is not None:
      kwargs[sig.name] = inference_params_value
    else:
      if sig_i == 0:
        # possible inlined first value
        value = serializer.deserialize(proto)
        if id(value) not in _ZERO_VALUE_IDS:
          # note missing values are not set to defaults, since they are not in parts
          # an actual zero value passed in must be set in an explicit part
          kwargs[sig.name] = value
        continue

      if sig.required or is_output:  # TODO allow optional outputs?
        raise ValueError(f'Missing required field: {sig.name}')
      continue
  if len(kwargs) == 1 and 'return' in kwargs:
    return kwargs['return']
  return kwargs


def get_stream_from_signature(signatures):
  '''
  Get the stream signature from the given signatures.
  '''
  for sig in signatures:
    if sig.iterator:
      return sig
  return None


def _is_empty_proto_data(data):
  if isinstance(data, np.ndarray):
    return False
  if isinstance(data, MessageProto):
    return not data.ByteSize()
  return not data


def _normalize_type(tp):
  '''
  Normalize the types for the given parameter.
  Returns the normalized type and whether the parameter is streaming.
  '''
  # stream type indicates streaming, not part of the data itself
  # it can only be used at the top-level of the var type
  streaming = (get_origin(tp) in [abc.Iterator, abc.Generator, abc.Iterable])
  if streaming:
    tp = get_args(tp)[0]

  return _normalize_data_type(tp), streaming


def _normalize_data_type(tp):
  # container types that need to be serialized as parts
  if get_origin(tp) == list and get_args(tp):
    return List[_normalize_data_type(get_args(tp)[0])]

  if get_origin(tp) == tuple:
    if not get_args(tp):
      raise TypeError('Tuple must have types specified')
    return Tuple[tuple(_normalize_data_type(val) for val in get_args(tp))]

  if isinstance(tp, (tuple, list)):
    return Tuple[tuple(_normalize_data_type(val) for val in tp)]

  if tp == data_types.NamedFields:
    raise TypeError('NamedFields must have types specified')

  # Handle dynamically generated NamedFields subclasses with annotations
  if isinstance(tp, type) and issubclass(tp, data_types.NamedFields) and hasattr(
      tp, '__annotations__'):
    return data_types.NamedFields(
        **{k: _normalize_data_type(v)
           for k, v in tp.__annotations__.items()})

  if isinstance(tp, (dict, data_types.NamedFields)):
    return data_types.NamedFields(**{name: _normalize_data_type(val) for name, val in tp.items()})

  # check if numpy array type, and if so, use ndarray
  if get_origin(tp) == np.ndarray:
    return np.ndarray

  # check for PIL images (sometimes types use the module, sometimes the class)
  # set these to use the Image data handler
  if tp in (data_types.Image, PIL.Image.Image):
    return data_types.Image

  if tp == PIL.Image:
    raise TypeError('Use PIL.Image.Image instead of PIL.Image module')

  # jsonable list and dict, these can be serialized as json
  # (tuple we want to keep as a tuple for args and returns, so don't include here)
  if tp in (list, dict, Dict) or (get_origin(tp) in (list, dict, Dict) and _is_jsonable(tp)):
    return data_types.JSON

  # check for known data types
  try:
    if tp in _DATA_TYPES:
      return tp
  except TypeError:
    pass  # not hashable type

  raise TypeError(f'Unsupported type: {tp}')


def _is_jsonable(tp):
  if tp in (dict, list, tuple, str, int, float, bool, type(None)):
    return True
  if get_origin(tp) in (tuple, list, dict):
    return all(_is_jsonable(val) for val in get_args(tp))
  return False


# type: name of the data type
# data_field: name of the field in the data proto
# serializer: serializer for the data type
_DataType = namedtuple('_DataType', ('type', 'serializer'))

_ZERO_VALUE_IDS = {id(None), id(''), id(b''), id(0), id(0.0), id(False)}

# simple, non-container types that correspond directly to a data field
_DATA_TYPES = {
    str:
        _DataType(resources_pb2.ModelTypeField.DataType.STR,
                  AtomicFieldSerializer('string_value')),
    bytes:
        _DataType(resources_pb2.ModelTypeField.DataType.BYTES,
                  AtomicFieldSerializer('bytes_value')),
    int:
        _DataType(resources_pb2.ModelTypeField.DataType.INT, AtomicFieldSerializer('int_value')),
    float:
        _DataType(resources_pb2.ModelTypeField.DataType.FLOAT,
                  AtomicFieldSerializer('float_value')),
    bool:
        _DataType(resources_pb2.ModelTypeField.DataType.BOOL, AtomicFieldSerializer('bool_value')),
    np.ndarray:
        _DataType(resources_pb2.ModelTypeField.DataType.NDARRAY, NDArraySerializer('ndarray')),
    data_types.JSON:
        _DataType(resources_pb2.ModelTypeField.DataType.JSON_DATA, JSONSerializer('string_value')
                 ),  # TODO change to json_value when new proto is ready
    data_types.Text:
        _DataType(resources_pb2.ModelTypeField.DataType.TEXT,
                  MessageSerializer('text', data_types.Text)),
    data_types.Image:
        _DataType(resources_pb2.ModelTypeField.DataType.IMAGE,
                  MessageSerializer('image', data_types.Image)),
    data_types.Concept:
        _DataType(resources_pb2.ModelTypeField.DataType.CONCEPT,
                  MessageSerializer('concepts', data_types.Concept)),
    data_types.Region:
        _DataType(resources_pb2.ModelTypeField.DataType.REGION,
                  MessageSerializer('regions', data_types.Region)),
    data_types.Frame:
        _DataType(resources_pb2.ModelTypeField.DataType.FRAME,
                  MessageSerializer('frames', data_types.Frame)),
    data_types.Audio:
        _DataType(resources_pb2.ModelTypeField.DataType.AUDIO,
                  MessageSerializer('audio', data_types.Audio)),
    data_types.Video:
        _DataType(resources_pb2.ModelTypeField.DataType.VIDEO,
                  MessageSerializer('video', data_types.Video)),
}

_SERIALIZERS_BY_TYPE_ENUM = {dt.type: dt.serializer for dt in _DATA_TYPES.values()}


class CompatibilitySerializer(Serializer):
  '''
  Serialization of basic value types, used for backwards compatibility
  with older models that don't have type signatures.
  '''

  def serialize(self, data_proto, value):
    tp = _normalize_data_type(type(value))

    try:
      serializer = _DATA_TYPES[tp].serializer
    except (KeyError, TypeError):
      raise TypeError(f'serializer currently only supports basic types, got {tp}')

    serializer.serialize(data_proto, value)

  def deserialize(self, data_proto):
    fields = [k.name for k, _ in data_proto.ListFields()]
    if 'parts' in fields:
      raise ValueError('serializer does not support parts')
    serializers = [
        serializer for serializer in _SERIALIZERS_BY_TYPE_ENUM.values()
        if serializer.field_name in fields
    ]
    if not serializers:
      raise ValueError('Returned data not recognized')
    if len(serializers) != 1:
      raise ValueError('Only single output supported for serializer')
    serializer = serializers[0]
    return serializer.deserialize(data_proto)
